#include <stdexcept>
#include <iomanip>
#include <tnt/ecpp.h>
#include <tnt/htmlescostream.h>
#include <tnt/httprequest.h>
#include <tnt/httpreply.h>
#include <openssl/md5.h>
#include "exception.h"
#include "live.h"
#include "setup.h"
#include "tools.h"


using namespace std;
using namespace tnt;

istream& operator>>( istream& is, tChannelID& ret )
{
	if ( is.rdbuf()->in_avail() > 0 ) {
		string line;
		if ( !getline( is, line ) || ( !line.empty() && !( ret = tChannelID::FromString( line.c_str() ) ).Valid() ) )
			is.setstate( ios::badbit );
	}
	return is;
}

namespace vdrlive {

string FormatDateTime( char const* format, time_t time )
{
	struct tm tm_r;
	if ( localtime_r( &time, &tm_r ) == 0 ) {
		ostringstream builder;
		builder << "cannot represent timestamp " << time << " as local time";
		throw runtime_error( builder.str() );
	}

	char result[ 256 ];
	if ( strftime( result, sizeof( result ), format, &tm_r ) == 0 ) {
		ostringstream builder;
		builder << "representation of timestamp " << time << " exceeds " << sizeof( result ) << " bytes";
		throw runtime_error( builder.str() );
	}
	return result;
}

string StringReplace( string const& text, string const& substring, string const& replacement )
{
	string result = text;
	string::size_type pos = 0;
	while ( ( pos = result.find( substring, pos ) ) != string::npos ) {
		result.replace( pos, substring.length(), replacement );
		pos += replacement.length();
	}
	return result;
}

vector< string > StringSplit( string const& text, char delimiter )
{
	vector< string > result;
	string::size_type last = 0, pos;
	while ( ( pos = text.find( delimiter, last ) ) != string::npos ) {
		result.push_back( text.substr( last, pos - last ) );
		last = pos + 1;
	}
	if ( last < text.length() )
		result.push_back( text.substr( last ) );
	return result;
}

int StringToInt( std::string const& string, int base )
{
	char* end;
	int result = strtol( string.c_str(), &end, base );
	if ( *end == '\0' )
		return result;
	return 0;
}

string StringRepeat(int times, const string& input)
{
	string result;
	for (int i = 0; i < times; i++) {
		result += input;
	}
	return result;
}

string StringWordTruncate(const string& input, size_t maxLen, bool& truncated)
{
	if (input.length() <= maxLen)
	{
		return input;
	}
	truncated = true;
	string result = input.substr(0, maxLen);
	size_t pos = result.find_last_of(" \t,;:.\n?!'\"/\\()[]{}*+-");
	return result.substr(0, pos);
}

string StringFormatBreak(string const& input)
{
	return StringReplace( input, "\n", "<br/>" );
}

string StringEscapeAndBreak( string const& input )
{
	stringstream plainBuilder;
	HtmlEscOstream builder( plainBuilder );
	builder << input;
	return StringReplace( plainBuilder.str(), "\n", "<br/>" );
}

string StringTrim(string const& str)
{
   string res = str;
   string::size_type pos = res.find_last_not_of(' ');
   if(pos != string::npos) {
      res.erase(pos + 1);
      pos = res.find_first_not_of(' ');
      if(pos != string::npos) res.erase(0, pos);
   }
   else res.erase(res.begin(), res.end());
   return res;
}

string ZeroPad(int number)
{
   ostringstream os;
   os << setw(2) << setfill('0') << number;
   return os.str();
}

std::string MD5Hash(std::string const& str)
{
	unsigned char md5[MD5_DIGEST_LENGTH];
	MD5(reinterpret_cast<const unsigned char*>(str.c_str()), str.size(), md5);

	ostringstream hashStr;
	hashStr << hex;
	for (size_t i = 0; i < MD5_DIGEST_LENGTH; i++)
		hashStr << (0 + md5[i]);

	return hashStr.str();
}

#define HOURS(x) ((x)/100)
#define MINUTES(x) ((x)%100)

std::string ExpandTimeString(std::string timestring)
{
	string::size_type colonpos = timestring.find(":");
	if (colonpos == string::npos)
	{
		if (timestring.size() == 1)
			timestring = "0" + timestring + ":00";
		else if (timestring.size() == 2)
			timestring = timestring + ":00";
		else if (timestring.size() == 3)
			timestring = "0" + string(timestring.begin(), timestring.begin() + 1) + ":" + string(timestring.begin() + 1, timestring.end());
		else
			timestring = string(timestring.begin(), timestring.begin() + 2) + ":" + string(timestring.begin() + 2, timestring.end());
	}
	else
	{
		string hours = string(timestring.begin(), timestring.begin() + colonpos);
		string mins = string(timestring.begin() + colonpos + 1, timestring.end());
		hours = string(std::max(0,(int)(2 - hours.size())), '0') + hours;
		mins = string(std::max(0,(int)(2 - mins.size())), '0') + mins;
		timestring = hours + ":" + mins;
	}
	return timestring;
}

time_t GetTimeT(std::string timestring) // timestring in HH:MM
{
	timestring = StringReplace(timestring, ":", "");
	int iTime = lexical_cast< int >( timestring );
	struct tm tm_r;
	time_t t = time(NULL);
	tm* tmnow = localtime_r(&t, &tm_r);
	tmnow->tm_hour = HOURS(iTime);
	tmnow->tm_min = MINUTES(iTime);
	return mktime(tmnow);
}

} // namespace vdrlive
